@@ -8,6 +8,7 @@ import (
 	"bytes"
 	"encoding/xml"
 	"fmt"
+	"log/slog"
 	"net/http"
 	"net/url"
 	"os"
@@ -17,146 +18,106 @@ import (
 	"strings"
 	"time"
 
-	"github.com/beevik/etree"            // BSD-2-clause
-	"github.com/gabriel-vasile/mimetype" // MIT License
-	"github.com/xmppo/go-xmpp"           // BSD-3-Clause
+	"github.com/ProtonMail/gopenpgp/v3/crypto" // MIT License
+	"github.com/beevik/etree"                  // BSD-2-clause
+	"github.com/gabriel-vasile/mimetype"       // MIT License
+	"github.com/xmppo/go-xmpp"                 // BSD-3-Clause
 )
 
-func httpUpload(client *xmpp.Client, iqc chan xmpp.IQ, jserver string, filePaths []string,
-	timeout time.Duration,
+func httpUpload(client *xmpp.Client, iqc chan xmpp.IQ, disc chan xmpp.DiscoItems, drc chan xmpp.DiscoResult,
+	jserver string, filePaths []string, timeout time.Duration, oxPrivKey *crypto.Key, keyRing *crypto.KeyRing,
 ) (urls []string, err error) {
 	var uploadComponent string
 	var maxFileSize int64
-	var iqDiscoItemsXMLQuery, iqDiscoInfoXMLQuery *etree.Element
 
 	// Query server for disco#items
-	iqContent, err := sendIQ(client, iqc, jserver, "get",
-		"<query xmlns='http://jabber.org/protocol/disco#items'/>")
+	slog.Info("http-upload: querying disco items", "server", jserver)
+	discoItems, err := getDiscoItems(client, disc, jserver)
 	if err != nil {
-		return urls, err
+		return
 	}
-	iqDiscoItemsXML := etree.NewDocument()
-	err = iqDiscoItemsXML.ReadFromBytes(iqContent.Query)
-	if err != nil {
-		return urls, err
-	}
-	iqDiscoItemsXMLQuery = iqDiscoItemsXML.SelectElement("query")
-	if iqDiscoItemsXMLQuery == nil {
-		return urls, fmt.Errorf("http-upload: no query element in disco items reply")
-	}
-	iqDiscoItemsXMLItems := iqDiscoItemsXMLQuery.SelectElements("item")
 
 	// Check the services reported by disco#items for the http upload service
-	for _, r := range iqDiscoItemsXMLItems {
-		jid := r.SelectAttr("jid")
-		iqDiscoInfoReqXML := etree.NewDocument()
-		iqDiscoInfoReqXML.WriteSettings.AttrSingleQuote = true
-		iqDiscoInfoReqXMLQuery := iqDiscoInfoReqXML.CreateElement("query")
-		iqDiscoInfoReqXMLQuery.CreateAttr("xmlns", nsDiscoInfo)
-		iqdi, err := iqDiscoInfoReqXML.WriteToString()
+	for _, r := range discoItems.Items {
+		slog.Info("http-upload: querying disco info", "disco-item", r.Jid)
+		discoResult, err := getDiscoInfo(client, drc, r.Jid)
 		if err != nil {
 			return urls, err
 		}
-		iqDiscoInfo, err := sendIQ(client, iqc, jid.Value, "get", iqdi)
-		if err != nil {
-			return urls, err
+		slog.Info("http-upload: checking identities for type = \"file\" and category = \"store\"")
+		for _, identity := range discoResult.Identities {
+			if identity.Type == "file" &&
+				identity.Category == "store" {
+				uploadComponent = r.Jid
+				slog.Info("http-upload:", "upload component", uploadComponent)
+				for _, x := range discoResult.X {
+					for i, field := range x.Field {
+						if field.Var == "max-file-size" {
+							if i > 0 {
+								if x.Field[i-1].Value[0] == nsHTTPUpload {
+									maxFileSize, err = strconv.ParseInt(field.Value[0], 10, 64)
+									if err != nil {
+										return urls, fmt.Errorf("http-upload: error while checking server maximum http upload file size")
+									}
+									slog.Info("http-upload:", "max file size", maxFileSize)
+									break
+								}
+							}
+						}
+					}
+				}
+				break
+			}
+			if uploadComponent != "" {
+				break
+			}
 		}
-		if iqDiscoInfo.Type != strResult {
-			continue
-		}
-		iqDiscoInfoXML := etree.NewDocument()
-		err = iqDiscoInfoXML.ReadFromBytes(iqDiscoInfo.Query)
-		if err != nil {
-			return urls, err
-		}
-		iqDiscoInfoXMLQuery = iqDiscoInfoXML.SelectElement("query")
-		if iqDiscoInfoXMLQuery == nil {
-			continue
-		}
-		iqDiscoInfoXMLIdentity := iqDiscoInfoXMLQuery.SelectElement("identity")
-		if iqDiscoInfoXMLIdentity == nil {
-			continue
-		}
-		iqDiscoInfoXMLType := iqDiscoInfoXMLIdentity.SelectAttr("type")
-		if iqDiscoInfoXMLType == nil {
-			continue
-		}
-		iqDiscoInfoXMLCategory := iqDiscoInfoXMLIdentity.SelectAttr("category")
-		if iqDiscoInfoXMLCategory == nil {
-			continue
-		}
-		if iqDiscoInfoXMLType.Value == "file" &&
-			iqDiscoInfoXMLCategory.Value == "store" {
-			uploadComponent = jid.Value
+		if uploadComponent != "" {
 			break
 		}
 	}
 	if uploadComponent == "" {
 		return urls, fmt.Errorf("http-upload: no http upload component found")
 	}
-	iqDiscoInfoXMLX := iqDiscoInfoXMLQuery.SelectElements("x")
-	for _, r := range iqDiscoInfoXMLX {
-		field := r.SelectElements("field")
-		for i, t := range field {
-			varAttr := t.SelectAttr("var")
-			if varAttr == nil {
-				continue
-			}
-			curFieldVal := t.SelectElement("value")
-			if curFieldVal == nil {
-				continue
-			}
-			if varAttr.Value == "max-file-size" {
-				var prevFieldVal *etree.Element
-				if i > 0 {
-					prevFieldVal = field[i-1].SelectElement("value")
-					if prevFieldVal == nil {
-						continue
-					}
-				}
-				if prevFieldVal.Text() == nsHTTPUpload {
-					maxFileSize, err = strconv.ParseInt(curFieldVal.Text(), 10, 64)
-					if err != nil {
-						return urls, fmt.Errorf("http-upload: error while checking server maximum http upload file size")
-					}
-				}
-			}
-		}
-	}
-
 	for _, filePath := range filePaths {
-		// Get file size
-		fileInfo, err := os.Stat(filePath)
-		if err != nil {
-			return urls, err
-		}
-		fileSize := fileInfo.Size()
-
 		// Read file
 		buffer, err := readFile(filePath)
 		if err != nil {
 			return urls, err
 		}
-
+		fileSize := buffer.Len()
+		slog.Info("http-upload: checking file size:", "file", filePath, "size", fileSize)
 		// Get mime type
 		mimeType := mimetype.Detect(buffer.Bytes()).String()
+		slog.Info("http-upload:", "mime type", mimeType)
 		var mimeTypeEscaped bytes.Buffer
 		xml.Escape(&mimeTypeEscaped, []byte(mimeType))
-
+		if oxPrivKey != nil {
+			buffer, err = legacyPGPEncryptFile(oxPrivKey, keyRing, buffer)
+			if err != nil {
+				return urls, err
+			}
+			fileSize = buffer.Len()
+			slog.Info("http-upload: updated file size after encrypting:", "file", filePath, "size", fileSize)
+		}
 		// Get file name
 		fileName := filepath.Base(filePath)
+		if oxPrivKey != nil {
+			fileName = getID() + ".pgp"
+		}
 		// Just use alphanumerical and some special characters for now
 		// to work around https://github.com/xmppo/go-xmpp/issues/132
 		reg := regexp.MustCompile(`[^a-zA-Z0-9\+\-\_\.]+`)
 		fileNameEscaped := reg.ReplaceAllString(fileName, "_")
+		slog.Info("http-upload: escape file", "name", fileNameEscaped)
 
 		// Check if the file size doesn't exceed the maximum file size of the http upload
 		// component if a maximum file size is reported, if not just continue and hope for
 		// the best.
 		if maxFileSize != 0 {
-			if fileSize > maxFileSize {
+			if int64(fileSize) > maxFileSize {
 				return urls, fmt.Errorf("http-upload: file size %s MiB is larger than the maximum file size allowed (%s MiB)",
-					strconv.FormatInt(fileSize/1024/1024, 10), strconv.FormatInt(maxFileSize/1024/1024, 10))
+					strconv.FormatInt(int64(fileSize)/1024/1024, 10), strconv.FormatInt(maxFileSize/1024/1024, 10))
 			}
 		}
 
@@ -173,6 +134,7 @@ func httpUpload(client *xmpp.Client, iqc chan xmpp.IQ, jserver string, filePaths
 		}
 
 		// Request http upload slot
+		slog.Info("http-upload: requesting upload slot")
 		uploadSlot, err := sendIQ(client, iqc, uploadComponent, "get", r)
 		if err != nil {
 			return urls, err
@@ -201,16 +163,19 @@ func httpUpload(client *xmpp.Client, iqc chan xmpp.IQ, jserver string, filePaths
 			return urls, fmt.Errorf("http-upload: upload slot does not provide https")
 		}
 		// Upload file
+		slog.Info("http-upload: uploading file to", "slot", iqHTTPUploadSlotXMLPutURL.Value)
 		httpTransport := &http.Transport{
 			IdleConnTimeout:     timeout,
 			TLSHandshakeTimeout: timeout,
 		}
 		proxyEnv := os.Getenv("HTTP_PROXY")
+		slog.Info("http-upload: getting environment variable:", "HTTP_PROXY", proxyEnv)
 		if proxyEnv != "" {
 			proxyURL, err := url.Parse(proxyEnv)
 			if err != nil {
 				return urls, err
 			}
+			slog.Info("http-upload: using http transport", "proxy", proxyURL)
 			httpTransport.Proxy = http.ProxyURL(proxyURL)
 		}
 		httpClient := &http.Client{Transport: httpTransport}
@@ -235,6 +200,7 @@ func httpUpload(client *xmpp.Client, iqc chan xmpp.IQ, jserver string, filePaths
 		if err != nil {
 			return urls, err
 		}
+		slog.Info("http-upload: received http", "status", resp.StatusCode)
 		// Test for http status code "200 OK" or "201 Created"
 		if resp.StatusCode != 200 && resp.StatusCode != 201 {
 			return urls, fmt.Errorf("http-upload: upload failed")
@@ -249,6 +215,7 @@ func httpUpload(client *xmpp.Client, iqc chan xmpp.IQ, jserver string, filePaths
 		if iqHTTPUploadSlotXMLGetURL == nil {
 			return urls, fmt.Errorf("http-upload: no url attribute")
 		}
+		slog.Info("http-upload: received get", "URL", iqHTTPUploadSlotXMLGetURL.Value)
 		err = resp.Body.Close()
 		if err != nil {
 			fmt.Println("http-upload: error while closing http request body:", err)
