@@ -2,6 +2,7 @@
 // Use of this source code is governed by the BSD-2-clause
 // license that can be found in the LICENSE file.
 
+// TODO: Add logging.
 package main
 
 import (
@@ -10,13 +11,73 @@ import (
 	"fmt"
 	"io"
 	"log"
+	"log/slog"
 	"runtime"
+	"strings"
 	"time"
 
 	"github.com/beevik/etree"  // BSD-2-clause
 	"github.com/xmppo/go-xmpp" // BSD-3-Clause
 )
 
+func getDiscoInfo(client *xmpp.Client, drc chan xmpp.DiscoResult, target string) (dr xmpp.DiscoResult, err error) {
+	id := getID()
+	c := make(chan xmpp.DiscoResult, defaultBufferSize)
+	go getDr(target, c, drc)
+	slog.Info("stanzahandling: requesting disco info")
+	_, err = client.RawInformation(client.JID(), target, id, "get",
+		fmt.Sprintf("<query xmlns='%s'/>", nsDiscoInfo))
+	if err != nil {
+		return
+	}
+	select {
+	case dr = <-c:
+		slog.Info("stanzahandling: requesting disco info: received reply")
+	case <-time.After(60 * time.Second):
+		slog.Info("stanzahandling: requesting disco info: timeout")
+		return dr, fmt.Errorf("get disco info: %s didn't reply to disco info request", target)
+	}
+	return
+}
+
+func getDr(jid string, c chan xmpp.DiscoResult, drc chan xmpp.DiscoResult) {
+	for {
+		dr := <-drc
+		c <- dr
+		return
+	}
+}
+
+func getDiscoItems(client *xmpp.Client, disc chan xmpp.DiscoItems, target string) (dis xmpp.DiscoItems, err error) {
+	id := getID()
+	c := make(chan xmpp.DiscoItems, defaultBufferSize)
+	go getDis(target, c, disc)
+	slog.Info("stanzahandling: requesting disco items")
+	_, err = client.RawInformation(client.JID(), target, id, "get",
+		fmt.Sprintf("<query xmlns='%s'/>", nsDiscoItems))
+	if err != nil {
+		return
+	}
+	select {
+	case dis = <-c:
+		slog.Info("stanzahandling: requesting disco items: received reply")
+	case <-time.After(60 * time.Second):
+		slog.Info("stanzahandling: requesting disco items: timeout")
+		return dis, fmt.Errorf("get disco items: %s didn't reply to disco items request", target)
+	}
+	return
+}
+
+func getDis(jid string, c chan xmpp.DiscoItems, disc chan xmpp.DiscoItems) {
+	for {
+		dis := <-disc
+		if dis.Jid == jid {
+			c <- dis
+			return
+		}
+	}
+}
+
 func sendIQ(client *xmpp.Client, iqc chan xmpp.IQ, target string, iQtype string, content string) (xmpp.IQ, error) {
 	var iq xmpp.IQ
 	id := getID()
@@ -25,12 +86,12 @@ func sendIQ(client *xmpp.Client, iqc chan xmpp.IQ, target string, iQtype string,
 	_, err := client.RawInformation(client.JID(), target, id,
 		iQtype, content)
 	if err != nil {
-		return iq, fmt.Errorf("sendIQ: failed to send iq: %w", err)
+		return iq, fmt.Errorf("send iq: failed to send iq: %w", err)
 	}
 	select {
 	case iq = <-c:
 	case <-time.After(60 * time.Second):
-		return iq, fmt.Errorf("sendIQ: server didn't reply to IQ: %s", content)
+		return iq, fmt.Errorf("send iq: server didn't reply to IQ: %s", content)
 	}
 	return iq, nil
 }
@@ -45,7 +106,17 @@ func getIQ(id string, c chan xmpp.IQ, iqc chan xmpp.IQ) {
 	}
 }
 
-func rcvStanzas(client *xmpp.Client, ctx context.Context, iqc chan xmpp.IQ, msgc chan xmpp.Chat) {
+func getPresence(from string, c chan xmpp.Presence, prsc chan xmpp.Presence) {
+	for {
+		p := <-prsc
+		if strings.HasPrefix(p.From, from) {
+			c <- p
+			return
+		}
+	}
+}
+
+func rcvStanzas(client *xmpp.Client, ctx context.Context, iqc chan xmpp.IQ, msgc chan xmpp.Chat, prsc chan xmpp.Presence, disc chan xmpp.DiscoItems, drc chan xmpp.DiscoResult) {
 	type messageError struct {
 		XMLName xml.Name `xml:"message-error"`
 		Text    string   `xml:"text"`
@@ -80,7 +151,7 @@ func rcvStanzas(client *xmpp.Client, ctx context.Context, iqc chan xmpp.IQ, msgc
 			case nil:
 				continue
 			default:
-				closeAndExit(client, fmt.Errorf("rcvStanzas: %v", err))
+				closeAndExit(client, fmt.Errorf("receive stanzas: %v", err))
 				return
 			}
 		case received = <-r:
@@ -97,6 +168,12 @@ func rcvStanzas(client *xmpp.Client, ctx context.Context, iqc chan xmpp.IQ, msgc
 				}
 			}
 			msgc <- v
+		case xmpp.Presence:
+			prsc <- v
+		case xmpp.DiscoItems:
+			disc <- v
+		case xmpp.DiscoResult:
+			drc <- v
 		case xmpp.IQ:
 			switch v.Type {
 			case "get":
